//+------------------------------------------------------------------+
//|                                           ActivationFunction.mqh |
//|                                  Copyright 2023, MetaQuotes Ltd. |
//|                                             https://www.mql5.com |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
vector VectorActivationFunction(const vector& A,const ENUM_ACTIVATION_FUNCTION af,const double param=0,const double param2=0,const double param3=0)
  {
   vector C(A.Size());

   if(af==AF_SOFTMAX)
     {
      double exp_sum=0;
      for(ulong i=0; i<A.Size(); i++)
        {
         C[i]=exp(A[i]);
         exp_sum+=C[i];
        }
      if(exp_sum==0)
         C.Fill(0.0);
      else
        {
         for(ulong i=0; i<A.Size(); i++)
            C[i]/=exp_sum;
        }
     }
   else
     {
      for(ulong i=0; i<A.Size(); i++)
         C[i]=ActivationFunction(af,A[i],param,param2,param3);
     }

   return(C);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
vector VectorDerivativeFunction(const vector& A,const ENUM_ACTIVATION_FUNCTION af,const double param=0,const double param2=0,const double param3=0)
  {
   vector C(A.Size());

   if(af==AF_SOFTMAX)
     {
      double exp_sum=0;
      for(ulong i=0; i<A.Size(); i++)
        {
         C[i]=exp(A[i]);
         exp_sum+=C[i];
        }
      if(exp_sum==0)
         C.Fill(0.0);
      else
        {
         for(ulong i=0; i<A.Size(); i++)
           {
            double y=C[i]/exp_sum;
            C[i]=y*(1-y);
           }
        }
     }
   else
     {
      for(ulong i=0; i<A.Size(); i++)
         C[i]=DerivativeFunction(af,A[i],param,param2,param3);
     }

   return(C);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActivationFunction(const ENUM_ACTIVATION_FUNCTION af,const double value,const double param=0,const double param2=0,const double param3=0)
  {
   double alpha,beta,max_value,treshold,theta;
//---
   switch(af)
     {
      case AF_NONE :
         return(value);

      case AF_ELU :
         if(value>=0)
            return(value);
         alpha=param;
         if(alpha<=0)
            alpha=1;
         return(alpha*(exp(value)-1));

      case AF_EXP :
         return(exp(value));

      case AF_GELU :
         return(0.5*value*(1+tanh(sqrt(M_2_PI)*(value+0.044715*value*value*value))));

      case AF_HARD_SIGMOID :
         if(value<-2.5)
            return(0);
         if(value>2.5)
            return(1);
         return(0.2*value+0.5);

      case AF_LINEAR :
         alpha=param;
         if(alpha==0)
            alpha=1;
         beta=param2;
         return(alpha*value+beta);

      case AF_LRELU :
         if(value>=0)
            return(value);
         alpha=param;
         if(alpha==0)
            alpha=0.3;
         return(alpha*value);

      case AF_RELU :
         alpha=param;
         if(alpha==0)
           {
            if(value>0)
               return(value);
            else
               return(0);
           }
         max_value=param2;
         if(value>=max_value)
            return(value);
         treshold=param3;
         return(alpha*(value-treshold));

      case AF_SELU :
         if(value>=0)
            return(1.05070098*value);
         return(1.05070098*1.67326324*(exp(value)-1));

      case AF_SIGMOID :
         return(1/(1+exp(-value)));

      case AF_SOFTPLUS :
         return(log(exp(value)+1));

      case AF_SOFTSIGN :
         return(value/(fabs(value)+1));

      case AF_SWISH :
         beta=param;
         return(value/(1+exp(-value*beta)));

      case AF_TANH :
         return(tanh(value));

      case AF_TRELU :
         theta=param;
         if(theta==0)
            theta=1;
         if(value>theta)
            return(value);
         return(0);

      case AF_PRELU :
         if(value>=0)
            return(value);
         alpha=param;
         return(alpha*value);
     }
//---
   return(0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double DerivativeFunction(const ENUM_ACTIVATION_FUNCTION af,const double value,const double param=0,const double param2=0,const double param3=0)
  {
   double alpha,value_3,max_value,theta,beta,tmp;
//---
   switch(af)
     {
      case AF_NONE :
         return(1);

      case AF_ELU :
         if(value>=0)
            return(1);
         alpha=param;
         if(alpha<=0)
            alpha=1;
         return(alpha*exp(value));

      case AF_EXP :
         return(exp(value));

      case AF_GELU :
         value_3=value*value*value;
         tmp=cosh(0.0356074*value_3+0.797885*value);
         return(0.5*tanh(0.0356774*value_3+0.398942*value)+(0.535161*value_3+0.398942*value)/(tmp*tmp)+0.5);

      case AF_HARD_SIGMOID :
         if(value<-2.5)
            return(0);
         if(value>2.5)
            return(0);
         return(0.2);

      case AF_LINEAR :
         alpha=param;
         if(alpha==0)
            alpha=1;
         return(alpha);

      case AF_LRELU :
         if(value>=0)
            return(1);
         alpha=param;
         if(alpha==0)
            alpha=0.3;
         return(alpha);

      case AF_RELU :
         alpha=param;
         if(alpha==0)
           {
            if(value>0)
               return(1);
            else
               return(0);
           }
         max_value=param2;
         if(value>=max_value)
            return(1);
         return(alpha);

      case AF_SELU :
         if(value>=0)
            return(1.05070098);
         return(1.05070098*1.67326324*exp(value));

      case AF_SIGMOID :
         tmp=exp(value);
         return(tmp/((tmp+1)*(tmp+1)));

      case AF_SOFTPLUS :
         tmp=exp(value);
         return(tmp/(tmp+1));

      case AF_SOFTSIGN :
         tmp=fabs(value)+1;
         return(1/(tmp*tmp));

      case AF_SWISH :
         beta=param;
         tmp=exp(beta*value);
         return((tmp*(beta*value+tmp+1)/((tmp+1)*(tmp+1))));

      case AF_TANH :
         tmp=cosh(value);
         return(1/(tmp*tmp));

      case AF_TRELU :
         theta=param;
         if(theta==0)
            theta=1;
         if(value>theta)
            return(1);
         return(0);

      case AF_PRELU :
         if(value>=0)
            return(1);
         alpha=param;
         return(alpha);
     }
//---
   return(0);
  }
//+------------------------------------------------------------------+
